package me.unurled.sacredrealms.sr.commands.player;

import static me.unurled.sacredrealms.sr.utils.Component.NOT_PLAYER;
import static me.unurled.sacredrealms.sr.utils.Component.NO_PERMISSION;
import static me.unurled.sacredrealms.sr.utils.Component.PLAYER_NOT_FOUND;
import static me.unurled.sacredrealms.sr.utils.Component.textComp;

import io.papermc.paper.command.brigadier.BasicCommand;
import io.papermc.paper.command.brigadier.CommandSourceStack;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import me.unurled.sacredrealms.sr.components.difficulty.Difficulty;
import me.unurled.sacredrealms.sr.components.player.PlayerManager;
import me.unurled.sacredrealms.sr.components.player.SRPlayer;
import me.unurled.sacredrealms.sr.gui.difficulty.DifficultyGUI;
import me.unurled.srcore.api.Manager;
import net.kyori.adventure.audience.Audience;
import net.kyori.adventure.text.Component;
import org.bukkit.Bukkit;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.jetbrains.annotations.NotNull;
import xyz.xenondevs.invui.window.Window;

/**
 * /difficulty ~open a gui for current player~ /difficulty [player] ~open a gui for another player~
 * /difficulty [player] [difficulty] ~set the difficulty for another player~
 */
public class DifficultyCommand implements BasicCommand {

  public static final String DIFFICULTY_SELF = "sr.difficulty.self";
  public static final String DIFFICULTY_MANAGE = "sr.difficulty.manage";
  public static final String CHOOSE_DIFFICULTY = "Choose Difficulty";

  public static void usage(CommandSender sender) {
    if (sender instanceof Player p) {
      if (p.hasPermission(DIFFICULTY_SELF)) {
        sender.sendMessage(textComp("<yellow>Usage: /difficulty"));
      }
      if (p.hasPermission(DIFFICULTY_MANAGE)) {
        sender.sendMessage(textComp("<yellow>Usage: /difficulty [player]"));
        sender.sendMessage(textComp("<yellow>Usage: /difficulty [player] [difficulty]"));
      }
    } else {
      sender.sendMessage(textComp("<yellow>Usage: /difficulty [player]"));
      sender.sendMessage(textComp("<yellow>Usage: /difficulty [player] [difficulty]"));
    }
  }

  private static void playerHandling(
      @NotNull CommandSender sender, @NotNull String[] args, @NotNull Player p) {
    if (!p.hasPermission(DIFFICULTY_SELF)) {
      sender.sendMessage(textComp(NO_PERMISSION));
      return;
    }
    if (args.length >= 1 && !p.hasPermission(DIFFICULTY_MANAGE)) {
      sender.sendMessage(textComp(NO_PERMISSION));
      usage(sender);
      return;
    }

    if (args.length == 0) {
      // open gui for sender
      Window window =
          Window.single()
              .setViewer(p)
              .setTitle(CHOOSE_DIFFICULTY)
              .setGui(DifficultyGUI.createGui())
              .build();
      window.open();
      return;
    }

    if (args.length == 1) {
      // open gui for target player
      Player target = p.getServer().getPlayer(args[0]);
      if (target == null) {
        sender.sendMessage(textComp(PLAYER_NOT_FOUND));
        return;
      }
      Window window =
          Window.single()
              .setViewer(p)
              .setTitle(CHOOSE_DIFFICULTY)
              .setGui(DifficultyGUI.createGui())
              .build();
      window.open();
      return;
    }

    if (args.length == 2) {
      // set difficulty for target player
      Player target = p.getServer().getPlayer(args[0]);
      if (target == null) {
        sender.sendMessage(textComp(PLAYER_NOT_FOUND));
        return;
      }
      // set difficulty
      PlayerManager pm = Manager.getInstance(PlayerManager.class);
      if (pm.isSRPlayer(target.getUniqueId())) {
        Difficulty difficulty = Difficulty.valueOf(args[1].toUpperCase());
        setDifficulty(sender, pm, target, difficulty);
      }
    }
  }

  private static void setDifficulty(
      @NotNull CommandSender sender,
      @NotNull PlayerManager pm,
      @NotNull Player target,
      Difficulty difficulty) {
    SRPlayer sr = pm.getPlayer(target.getUniqueId());
    if (sr == null) {
      sender.sendMessage(textComp(PLAYER_NOT_FOUND));
      return;
    }
    sr.setDifficulty(difficulty);
    Audience.audience(target)
        .sendMessage(
            Component.text("Your difficulty has been set to ")
                .append(difficulty.getName())
                .append(textComp("."))
                .color(difficulty.getColor()));
  }

  private static void consoleHandler(
      @NotNull CommandSender sender, @NotNull String @NotNull [] args) {
    if (args.length == 0) {
      sender.sendMessage(textComp(NOT_PLAYER));
      usage(sender);
      return;
    }

    Player target = Bukkit.getPlayer(args[0]);
    if (target == null) {
      sender.sendMessage(textComp(PLAYER_NOT_FOUND));
      return;
    }
    if (args.length == 1) {
      Window window =
          Window.single()
              .setViewer(target)
              .setTitle(CHOOSE_DIFFICULTY)
              .setGui(DifficultyGUI.createGui())
              .build();
      window.open();
    } else {
      Difficulty difficulty;
      try {
        difficulty = Difficulty.valueOf(args[1].toUpperCase());
      } catch (IllegalArgumentException e) {
        sender.sendMessage(textComp("<red>Invalid difficulty."));
        return;
      }
      PlayerManager pm = Manager.getInstance(PlayerManager.class);
      if (pm.isSRPlayer(target.getUniqueId())) {
        setDifficulty(sender, pm, target, difficulty);
      }
    }
  }

  /**
   * Executes the command with the given {@link CommandSourceStack} and arguments.
   *
   * @param commandSourceStack the commandSourceStack of the command
   * @param args the arguments of the command ignoring repeated spaces
   */
  @Override
  public void execute(@NotNull CommandSourceStack commandSourceStack, @NotNull String[] args) {
    CommandSender sender = commandSourceStack.getSender();
    if (sender instanceof Player p) {
      playerHandling(sender, args, p);
    } else {
      consoleHandler(sender, args);
    }
  }

  /**
   * Suggests possible completions for the given command {@link CommandSourceStack} and arguments.
   *
   * @param commandSourceStack the commandSourceStack of the command
   * @param args the arguments of the command including repeated spaces
   * @return a collection of suggestions
   */
  @Override
  public @NotNull Collection<String> suggest(
      @NotNull CommandSourceStack commandSourceStack, @NotNull String[] args) {
    CommandSender sender = commandSourceStack.getSender();
    if (args.length == 0 && sender.hasPermission(DIFFICULTY_MANAGE)) {
      return Bukkit.getOnlinePlayers().stream().map(Player::getName).toList();
    }
    if (args.length == 1 && sender.hasPermission(DIFFICULTY_MANAGE)) {
      return Arrays.stream(Difficulty.values()).map(Enum::name).toList();
    }
    return List.of();
  }
}
