package me.unurled.sacredrealms.sr.components.player;

import static me.unurled.sacredrealms.sr.utils.Component.textComp;
import static me.unurled.sacredrealms.sr.utils.Logger.error;
import static me.unurled.sacredrealms.sr.utils.Logger.log;
import static me.unurled.sacredrealms.sr.utils.SRPlayerUtils.updateActionBar;

import com.destroystokyo.paper.event.player.PlayerArmorChangeEvent;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.net.URI;
import java.util.HashMap;
import java.util.UUID;
import me.unurled.sacredrealms.sr.SR;
import me.unurled.sacredrealms.sr.components.pack.ResourcePackManager;
import me.unurled.sacredrealms.sr.data.DataHandler;
import me.unurled.sacredrealms.sr.data.DataManager;
import me.unurled.sacredrealms.sr.events.player.PlayerLevelUpEvent;
import me.unurled.sacredrealms.sr.utils.Items;
import me.unurled.srcore.api.Manager;
import net.kyori.adventure.resource.ResourcePackInfo;
import net.kyori.adventure.resource.ResourcePackRequest;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.entity.FoodLevelChangeEvent;
import org.bukkit.event.player.PlayerItemHeldEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.event.player.PlayerQuitEvent;
import org.bukkit.event.player.PlayerResourcePackStatusEvent.Status;
import org.bukkit.inventory.Inventory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.potion.PotionEffect;
import org.bukkit.scheduler.BukkitTask;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class PlayerManager extends Manager {

  public static final String PLAYER_KEY = "sr.players.";
  public static final String DATA_MANAGER_NULL = "DataManager is null, Can't save player data.";
  private final HashMap<UUID, SRPlayer> players;
  private final HashMap<UUID, BukkitTask> actionBarTasks;
  private DataHandler dh;

  public PlayerManager() {
    super();
    players = new HashMap<>();
    actionBarTasks = new HashMap<>();
  }

  /** Save the data */
  @Override
  public void saveData() {
    if (dh == null) {
      DataManager dm = Manager.getInstance(DataManager.class);
      if (dm == null) {
        error(DATA_MANAGER_NULL);
        return;
      }
      dh = dm.getDataHandler();
    }

    // save all players
    for (SRPlayer player : players.values()) {
      savePlayer(player);
    }

    players.clear();
  }

  @Nullable
  public SRPlayer getPlayer(UUID uuid) {
    if (players.containsKey(uuid)) {
      return players.get(uuid);
    }

    return loadPlayer(uuid);
  }

  public void addPlayer(@NotNull SRPlayer player) {
    players.put(player.getUuid(), player);
  }

  public void removePlayer(UUID uuid) {
    players.remove(uuid);
  }

  public boolean isSRPlayer(UUID uuid) {
    return players.containsKey(uuid);
  }

  public void savePlayer(SRPlayer player) {
    if (dh == null) {
      DataManager dm = Manager.getInstance(DataManager.class);
      if (dm == null) {
        error(DATA_MANAGER_NULL);
        return;
      }
      dh = dm.getDataHandler();
    }
    // save all players
    Gson gson =
        new GsonBuilder()
            .excludeFieldsWithoutExposeAnnotation()
            .registerTypeAdapter(Inventory.class, new InventorySerializer())
            .registerTypeAdapter(PotionEffect.class, new PotionEffectSerializer())
            .create();
    dh.set(PLAYER_KEY + player.getUuid(), gson.toJson(player));
  }

  @EventHandler
  public void onPlayerJoin(@NotNull PlayerJoinEvent e) {
    SRPlayer srPlayer = loadPlayer(e.getPlayer().getUniqueId());
    if (srPlayer == null) {
      return;
    }
    if (srPlayer.isFirstTime()) {
      e.joinMessage(
          textComp(
              "<gold>"
                  + e.getPlayer().getName()
                  + " has joined for "
                  + "the first time!<newline><gold>Welcome to "
                  + "the "
                  + "server!"));
    } else {
      e.joinMessage(
          textComp(
              "<gray>[</gray><green>+</green><gray>]</gray> <white>" + e.getPlayer().getName()));
    }
    addPlayer(srPlayer);

    if (srPlayer.getInventory() != null) {
      e.getPlayer().getInventory().clear();
      e.getPlayer().getInventory().setContents(srPlayer.getInventory().getContents());
    }

    // TODO: does not update when player health has changed for example
    BukkitTask task =
        Bukkit.getScheduler()
            .runTaskTimerAsynchronously(
                SR.getPlugin(), () -> updateActionBar(e.getPlayer()), 0, 20L);

    actionBarTasks.put(e.getPlayer().getUniqueId(), task);
    e.getPlayer().updateInventory();

    // player resource pack give
    ResourcePackManager rpm = Manager.getInstance(ResourcePackManager.class);
    if (rpm == null) {
      error("ResourcePackManager is null, Can't give player resource pack.");
      return;
    }
    try {
      URI uri = new URI(rpm.getHostingProvider().getMinecraftPackURL());
      ResourcePackRequest request =
          ResourcePackRequest.addingRequest(
              ResourcePackInfo.resourcePackInfo(
                  rpm.getHostingProvider().getPackUUID(),
                  uri,
                  rpm.getHostingProvider().getOriginalSHA1()));
      e.getPlayer().sendResourcePacks(request);
      if (e.getPlayer().getResourcePackStatus() != null
          && !e.getPlayer().getResourcePackStatus().equals(Status.ACCEPTED)) {
        e.getPlayer().kick(textComp("You must accept the resource pack."));
      }
    } catch (Exception ex) {
      error("Failed to send resource pack to player. " + ex.getMessage());
      log(rpm.getHostingProvider().getOriginalSHA1());
      log(rpm.getHostingProvider().getPackUUID().toString());
      log(rpm.getHostingProvider().getMinecraftPackURL());
    }
  }

  @Nullable
  private SRPlayer loadPlayer(UUID p) {
    if (dh == null) {
      DataManager dm = Manager.getInstance(DataManager.class);
      if (dm == null) {
        error(DATA_MANAGER_NULL);
        return null;
      }
      dh = dm.getDataHandler();
    }

    Gson gson =
        new GsonBuilder()
            .registerTypeAdapter(Inventory.class, new InventoryDeserializer())
            .registerTypeAdapter(PotionEffect.class, new PotionEffectDeserializer())
            .excludeFieldsWithoutExposeAnnotation()
            .create();
    String json = dh.get(PLAYER_KEY + p);
    SRPlayer srPlayer = gson.fromJson(json, SRPlayer.class);

    if (srPlayer == null) {
      srPlayer = new SRPlayer(p);
    } else {
      srPlayer.setFirstTime(false);
    }
    return srPlayer;
  }

  @EventHandler
  public void onPlayerQuit(@NotNull PlayerQuitEvent e) {
    SRPlayer player = getPlayer(e.getPlayer().getUniqueId());
    if (player == null) {
      return;
    }
    // save for player
    savePlayer(player);

    if (actionBarTasks.containsKey(e.getPlayer().getUniqueId())) {
      actionBarTasks.get(e.getPlayer().getUniqueId()).cancel();
      actionBarTasks.remove(e.getPlayer().getUniqueId());
    }

    removePlayer(e.getPlayer().getUniqueId());
    Bukkit.getServer()
        .broadcast(textComp("<gray>[<red>-<gray>] <white>" + e.getPlayer().getName()));
  }

  @EventHandler
  public void onPlayerInventorySlotChangeEvent(@NotNull PlayerItemHeldEvent e) {
    SRPlayer player = getPlayer(e.getPlayer().getUniqueId());
    if (player == null) {
      return;
    }
    player.setInventory(e.getPlayer().getInventory());

    ItemStack oldItem = e.getPlayer().getInventory().getItem(e.getPreviousSlot());
    ItemStack newItem = e.getPlayer().getInventory().getItem(e.getNewSlot());

    Items.updatePlayer(e.getPlayer(), oldItem, newItem);
  }

  @EventHandler
  public void onPlayerArmorChange(@NotNull PlayerArmorChangeEvent e) {
    SRPlayer player = getPlayer(e.getPlayer().getUniqueId());
    if (player == null) {
      return;
    }
    player.setInventory(e.getPlayer().getInventory());

    Items.updatePlayer(e.getPlayer(), e.getOldItem(), e.getNewItem());
  }

  @EventHandler
  public void onHunger(@NotNull FoodLevelChangeEvent e) {
    if (e.getEntity() instanceof Player) {
      SRPlayer player = getPlayer(e.getEntity().getUniqueId());
      if (player == null) {
        return;
      }
      e.setCancelled(true);
    }
  }

  public void levelUp(SRPlayer p, int previousLevel) {
    PlayerLevelUpEvent event = new PlayerLevelUpEvent(p, previousLevel);
    Bukkit.getScheduler().runTaskAsynchronously(SR.getPlugin(), event::callEvent);
    Player player = Bukkit.getPlayer(p.getUuid());
    if (player != null) {
      player.sendMessage(event.getMessage());
    }
  }
}
