package me.unurled.sacredrealms.sr.utils;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.List;

public class HTTPRequestMultipartBody {
  private final byte[] bytes;
  private String boundary;

  private HTTPRequestMultipartBody(byte[] bytes, String boundary) {
    this.bytes = bytes;
    this.boundary = boundary;
  }

  public String getBoundary() {
    return boundary;
  }

  public void setBoundary(String boundary) {
    this.boundary = boundary;
  }

  public String getContentType() {
    return "multipart/form-data; boundary=" + this.getBoundary();
  }

  public byte[] getBody() {
    return this.bytes;
  }

  public static class Builder {

    public static final String CONTENT_TYPE = "Content-Type: application/octet-stream\r\n\r\n";
    List<MultiPartRecord> parts;

    public Builder() {
      this.parts = new ArrayList<>();
    }

    public Builder addPart(String fieldName, String fieldValue) {
      MultiPartRecord part = new MultiPartRecord();
      part.setFieldName(fieldName);
      part.setContent(fieldValue);
      String defaultMimetype = "text/plain";
      part.setContentType(defaultMimetype);
      this.parts.add(part);
      return this;
    }

    public Builder addPart(String fieldName, String fieldValue, String contentType) {
      MultiPartRecord part = new MultiPartRecord();
      part.setFieldName(fieldName);
      part.setContent(fieldValue);
      part.setContentType(contentType);
      this.parts.add(part);
      return this;
    }

    public Builder addPart(
        String fieldName, Object fieldValue, String contentType, String fileName) {
      MultiPartRecord part = new MultiPartRecord();
      part.setFieldName(fieldName);
      part.setContent(fieldValue);
      part.setContentType(contentType);
      part.setFilename(fileName);
      this.parts.add(part);
      return this;
    }

    public HTTPRequestMultipartBody build() throws IOException {
      String boundary = new BigInteger(256, new SecureRandom()).toString();
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      for (MultiPartRecord multiPartRecord : parts) {
        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder
            .append("--")
            .append(boundary)
            .append("\r\n")
            .append("Content-Disposition: form-data; name=\"")
            .append(multiPartRecord.getFieldName());
        if (multiPartRecord.getFilename() != null) {
          stringBuilder.append("\"; filename=\"").append(multiPartRecord.getFilename());
        }
        out.write(stringBuilder.toString().getBytes(StandardCharsets.UTF_8));
        out.write(("\"\r\n").getBytes(StandardCharsets.UTF_8));
        Object content = multiPartRecord.getContent();
        if (content instanceof String cont) {
          out.write(("\r\n\r\n").getBytes(StandardCharsets.UTF_8));
          out.write(cont.getBytes(StandardCharsets.UTF_8));
        } else if (content instanceof byte[] cont) {
          out.write(CONTENT_TYPE.getBytes(StandardCharsets.UTF_8));
          out.write(cont);
        } else if (content instanceof File file) {
          out.write(CONTENT_TYPE.getBytes(StandardCharsets.UTF_8));
          Files.copy(file.toPath(), out);
        } else {
          out.write(CONTENT_TYPE.getBytes(StandardCharsets.UTF_8));
          ObjectOutputStream objectOutputStream = new ObjectOutputStream(out);
          objectOutputStream.writeObject(content);
          objectOutputStream.flush();
        }
        out.write("\r\n".getBytes(StandardCharsets.UTF_8));
      }
      out.write(("--" + boundary + "--\r\n").getBytes(StandardCharsets.UTF_8));

      return new HTTPRequestMultipartBody(out.toByteArray(), boundary);
    }

    public static class MultiPartRecord {
      private String fieldName;
      private String filename;
      private String contentType;
      private Object content;

      public String getFieldName() {
        return fieldName;
      }

      public void setFieldName(String fieldName) {
        this.fieldName = fieldName;
      }

      public String getFilename() {
        return filename;
      }

      public void setFilename(String filename) {
        this.filename = filename;
      }

      public String getContentType() {
        return contentType;
      }

      public void setContentType(String contentType) {
        this.contentType = contentType;
      }

      public Object getContent() {
        return content;
      }

      public void setContent(Object content) {
        this.content = content;
      }
    }
  }
}
